"""
tmdsclient is a fully typed, async client for the TMDS API.
The client is _not_ autogenerated.
Find the TMDS Open API description here: https://techmasterdata.xtk-dev.de/index.html
"""

from .client.config import TmdsConfig

# convenience exports
from .client.tmdsclient import TmdsClient
